#  Copyright (c) 2023. Lena "Teekeks" During <info@teawork.de>
"""
Objects used by this Library
----------------------------

.. toctree::
   :hidden:
   :maxdepth: 1

   twitchAPI.object.base
   twitchAPI.object.api
   twitchAPI.object.eventsub

.. autosummary::

   base
   api
   eventsub

"""

__all__ = []

